package com.richards.connector;

import java.io.File;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.concurrent.atomic.AtomicInteger;

import javax.sql.rowset.serial.SerialBlob;

import org.apache.log4j.Logger;
import org.apache.log4j.xml.DOMConfigurator;

public class MessageInsert {
	private Connector connector;
	private Thread batchExecuteThread;
	private Thread requestGenerateThread;
	private Thread responseGenerateThread;
	private AtomicInteger insertCount;
	private Statement responseStatement;
	private PreparedStatement requestStatement;
	private Logger logger = Logger.getLogger(MessageInsert.class);
	private Worker worker;
	private Connection connection;
	private String requestQuery;
	private RequestInsertThread requestInsertThread;
	private ResponseInsertThread responseInsertThread;

	public MessageInsert() throws Exception {
		insertCount = new AtomicInteger(0);
		connector = Connector.getConnector();
		connection = connector.getConnection();
		connection.setAutoCommit(false);
		requestQuery = "insert into request (requestTimeStamp, clientSeqID, systemSeqID, sendingTimeStamp, clientID, supplierID, clientAddress, destinationAddress,msgProp1, msgProp2, msgProp3, msgProp4, msgProp5, msgProp6, msgProp7, msgProp8,retryAttempt, msg)values(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
		requestStatement = connection.prepareStatement(requestQuery);
		responseStatement = connection.createStatement();

		worker = new Worker();
		batchExecuteThread = new Thread(worker);
		batchExecuteThread.start();

		requestInsertThread = new RequestInsertThread(10000);
		requestGenerateThread = new Thread(requestInsertThread);
		requestGenerateThread.start();

		responseInsertThread = new ResponseInsertThread(10000);
		responseGenerateThread = new Thread(responseInsertThread);
		responseGenerateThread.start();
		insertCount = new AtomicInteger(0);
	}

	public void saveResponse(int clientID, int systemSeqID, int supplierID,
			long responseTimeStamp, String msgID, int status) {
		String query = null;
		try {
			query = "insert into response(clientID,systemSeqID,supplierID,responseTimeStamp,msgID,status) values("
					+ clientID
					+ ","
					+ systemSeqID
					+ ","
					+ supplierID
					+ ",'"
					+ new Timestamp(responseTimeStamp)
					+ "'"
					+ ",'"
					+ msgID
					+ "'" + "," + status + ")";
			responseStatement.addBatch(query);
			logger.info("Added response to the batch successfully");
			// insertCount.incrementAndGet();
			if (insertCount.incrementAndGet() > connector.getBatchSize()) {
				saveBatch();
			}
		} catch (SQLException e) {
			logger.fatal(
					"Exception while adding the response query into the batch. Query:"
							+ query, e);
		}
	}

	public void saveRequest(long requestTimeStamp, int clientSeqID,
			int systemSeqID, long sendingTimeStamp, int clientID,
			int supplierID, String clientAddress, String destinationAddress,
			byte msgProp1, byte msgProp2, byte msgProp3, byte msgProp4,
			byte msgProp5, byte msgProp6, String msgProp7, byte msgProp8,
			byte[] msg) {
		try {
			requestStatement.setTimestamp(1, new Timestamp(requestTimeStamp));
			requestStatement.setInt(2, clientSeqID);
			requestStatement.setInt(3, systemSeqID);
			requestStatement.setTimestamp(4, new Timestamp(sendingTimeStamp));
			requestStatement.setInt(5, clientID);
			requestStatement.setInt(6, supplierID);
			requestStatement.setString(7, clientAddress);
			requestStatement.setString(8, destinationAddress);
			requestStatement.setByte(9, msgProp1);
			requestStatement.setByte(10, msgProp2);
			requestStatement.setByte(11, msgProp3);
			requestStatement.setByte(12, msgProp4);
			requestStatement.setByte(13, msgProp5);
			requestStatement.setByte(14, msgProp6);
			requestStatement.setString(15, msgProp7);
			requestStatement.setByte(16, msgProp8);
			requestStatement.setInt(17, 0);
			requestStatement.setBlob(18, new SerialBlob(msg));
			requestStatement.addBatch();
			logger.info("Added request message to the batch successfully");
			if (insertCount.incrementAndGet() > connector.getBatchSize()) {
				saveBatch();
			}
		} catch (SQLException e) {
			// for the moment i am avoiding the byte[] representation.
			logger.fatal("Exception while executing " + requestQuery
					+ " values " + requestTimeStamp + ", " + clientSeqID + ", "
					+ systemSeqID + ", " + sendingTimeStamp + ", " + clientID
					+ ", " + supplierID + ", " + clientAddress + ", "
					+ destinationAddress + ", " + msgProp1 + ", " + msgProp2
					+ ", " + msgProp3 + ", " + msgProp4 + ", " + msgProp5
					+ ", " + msgProp6 + ", " + msgProp7 + ", " + msgProp8);
		}
	}

	private class Worker implements Runnable {
		private boolean loop = true;

		public boolean isLoop() {
			return loop;
		}

		public void setLoop(boolean loop) {
			this.loop = loop;
		}

		@Override
		public void run() {

			while (loop) {
				saveBatch();
				try {
					Thread.sleep(connector.getBatchTimeout());
				} catch (InterruptedException e) {
					logger
							.info("Exception while sleeping the Worket thread",
									e);
				}
			}
		}
	}

	private class ResponseInsertThread implements Runnable {
		private int recordCount;
		private boolean loop = true;

		public void setLoop(boolean loop) {
			this.loop = loop;
		}

		public ResponseInsertThread(int recordCount) {
			super();
			this.recordCount = recordCount;
		}

		@Override
		public void run() {
			for (int i = 0; i < recordCount && loop; i++) {
				saveResponse(2, i, 4, System.currentTimeMillis(), "msg" + i,
						200);
			}

		}

	}

	private class RequestInsertThread implements Runnable {
		private int recordCount;
		private boolean loop = true;

		public void setLoop(boolean loop) {
			this.loop = loop;
		}

		public RequestInsertThread(int requestCount) {
			this.recordCount = requestCount;
		}

		@Override
		public void run() {
			byte[] array = "qwertyuioplkjhgfdsazxcvbnm1123456677880poiuytiheqjehkqwheqhwehqkesdmnfbsfbksjdfhsdsfjsp"
					.getBytes();
			for (int i = 0; i < recordCount && loop; i++) {
				saveRequest(System.currentTimeMillis(), i, i + 1, System
						.currentTimeMillis(), 2, 3, "indddd", " ktmmmm",
						(byte) 0x1, (byte) 0x01, (byte) 0x01, (byte) 0x01,
						(byte) 0x01, (byte) 0x01, "12345", (byte) 0x01, array);
			}

		}
	}

	public synchronized void saveBatch() {
		int batchCount = 0;
		try {
			batchCount = requestStatement.executeBatch().length;
		} catch (SQLException e) {
			logger.fatal("Exception while executing batch of requestStatement",
					e);
		}
		try {
			batchCount += responseStatement.executeBatch().length;
		} catch (SQLException e) {
			logger.fatal("Exception while executing batch of requestStatement",
					e);
		}
		try {
			connection.commit();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			logger.info("Exception while commmiting the request and responses",
					e);
		}
		if (batchCount > 0) {
			logger.info("Batch Insert Count:" + batchCount);
		}
		insertCount.set(0);
	}

	public void closeResources() {
		if (requestInsertThread != null) {
			requestInsertThread.setLoop(false);
			requestInsertThread = null;
		}
		if (requestGenerateThread != null) {
			requestGenerateThread = null;
		}

		if (responseInsertThread != null) {
			responseInsertThread.setLoop(false);
			responseInsertThread = null;
		}
		if (responseGenerateThread != null) {
			responseGenerateThread = null;
		}

		// call a save batch here to make sure that before the resources are
		// closed whatever is pending is saved into the database.

		saveBatch();

		if (requestStatement != null) {
			try {
				requestStatement.close();
			} catch (Exception e) {
				logger.fatal(
						"Exception while attempting to close requestStatment",
						e);
			} finally {
				requestStatement = null;
			}
		}
		if (responseStatement != null) {
			try {
				responseStatement.close();
			} catch (Exception e) {
				logger.fatal("Exception while closing the responseStatment", e);
			} finally {
				responseStatement = null;
			}
		}
		if (connection != null) {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.fatal("Exception while closing the connection");
			} finally {
				connection = null;
			}
		}
		if (worker != null) {
			worker.setLoop(false);
			try {
				Thread.sleep(5000L);
			} catch (Exception e) {
				logger.fatal("Exception while sleeping in closeResources", e);
			}
			worker = null;
		}

		if (batchExecuteThread != null) {
			batchExecuteThread = null;
		}
	}

	public static void main(String args[]) {
		MessageInsert insert = null;
		DOMConfigurator.configure("." + File.separator + "config"
				+ File.separator + "log4j.xml");
		Runtime.getRuntime().addShutdownHook(new ShutdownThread(insert));
		try {
			insert = new MessageInsert();

		} catch (Exception e) {
			Logger.getLogger(MessageInsert.class).fatal(
					"Exception while creating insert object", e);
			System.exit(0);
		}
	}

	private static class ShutdownThread extends Thread {
		private MessageInsert insert;

		public ShutdownThread(MessageInsert insert) {
			this.insert = insert;
		}

		public void run() {
			if (insert != null) {
				insert.closeResources();
			}
		}
	}
}
